/// <reference path="typings/index.d.ts" />
export interface EtcdKV {
    /** key is the key in bytes. An empty key is not allowed. */
    key: Buffer;
    /** create_revision is the revision of last creation on this key. */
    create_revision: number;
    /** mod_revision is the revision of last modification on this key. */
    mod_revision: number;
    /**
     * version is the version of the key. A deletion resets
     * the version to zero and any modification of the key
     * increases its version.
     */
    version: number;
    /** value is the value held by the key, in bytes. */
    value: Buffer;
    /**
     * lease is the ID of the lease that attached to key.
     * When the attached lease expires, the key will be deleted.
     * If lease is 0, then no lease is attached to the key.
     */
    lease: string;
}
export interface EtcdOptions {
    /** TTL of the managed client lease. seconds */
    appLeaseTtl?: number;
    /** Internal in which to keep alive the lease. milleseconds */
    appLeaseKeepAlive?: number;
}
export declare class Etcd {
    static defaults: EtcdOptions;
    servers: string[];
    options: EtcdOptions;
    credentials: any;
    clients: any;
    clientLease: string;
    clientLeasePromise: Promise<string>;
    constructor(servers?: string[], options?: EtcdOptions);
    private getBuffer(obj);
    private callClient(client, method, arg);
    private callClientStream(client, method);
    getClientLease(): Promise<string>;
    /**
     * blocking version of get()
     * @see {@link get}
     */
    getSync(key: string, returnType?: "value"): string;
    getSync(key: string, returnType: "json"): any;
    getSync(key: string, returnType: "buffer"): Buffer;
    getSync(key: string, returnType: "raw"): EtcdKV;
    /**
     * get a key from etcd
     * @param key - key to get
     * @return value of the key
     */
    get(key: string, returnType?: "value"): Promise<string>;
    get(key: string, returnType: "json"): Promise<any>;
    get(key: string, returnType: "buffer"): Promise<Buffer>;
    get(key: string, returnType: "raw"): Promise<EtcdKV>;
    /**
     * blocking version of range()
     * @see {@link range}
     */
    rangeSync(fromKey: string, toKey: string, returnType?: "value" | "json" | "buffer" | "raw"): any;
    /**
     * get a range of key from etcd
     * @param fromKey - key to start from
     * @param toKey - key to end on
     * @return value of the key
     */
    range(fromKey: string, toKey: string, returnType?: "value" | "json" | "buffer" | "raw"): Promise<any>;
    /**
     * blocking version of set()
     * @see {@link set}
     */
    setSync(key: string, value: any, lease?: string): boolean;
    /**
     * set a key/value to etcd.
     * * a `Buffer` is saved as-is
     * * `null` is saved as an empty string
     * * `undefined` is saved as an empty string
     * * strings are saved as-is
     * * everthing else is tried to `JSON.stringify(value)`
     *
     * if you set "lease" to "client" a lease uniq to the instance of the Client will be used. This lease
     * will be created automatic and will be kept alive automatic.
     *
     * @param key - etcd key to set
     * @param value - content to set
     * @param lease - lease ID to use, or `client` to the client lease
     * @return `true`
     */
    set(key: string, value: any, lease?: string): Promise<boolean>;
    /**
     * blocking version of delete()
     * @see {@link delete}
     */
    deleteSync(key: string, keyTo?: string): number;
    /**
     * delete a key from etcd.
     * @param key - etcd key to delete
     * @return number of keys deleted
     */
    delete(key: string, keyTo?: string): Promise<number>;
    /**
     * blocking version of leaseGrant()
     * @see {@link leaseGrant}
     */
    leaseGrantSync(ttl: number): string;
    /**
     * request a new lease from etcd
     * @param ttl - requested ttl of the lease
     * @return
     */
    leaseGrant(ttl?: number): Promise<string>;
    /**
     * keep a lease alive
     * @param lease - lease to keep alive
     * @param interval - interval in which to send keep alives
     * @return id of interval used
     */
    leaseKeepAlive(lease: string, interval?: number): number;
}
